# -*- coding: utf-8 -*-
"""Pro Custodibus Agent."""

__version__ = "1.7.1"

DEFAULT_API_URL = "https://api.custodib.us"
DEFAULT_APP_URL = "https://pro.custodib.us"
DOCS_URL = "https://docs.procustodibus.com"
